/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;



/**
 * base reopsurce for a BodyDataSink/Source
 *   
 * @author grro@xlightweb.org
 */
abstract class AbstractResource  {
	
    private static final Logger LOG = Logger.getLogger(AbstractResource.class.getName());
    
    
	private final ArrayList<IBodyCloseListener> closeListeners = new ArrayList<IBodyCloseListener>();
	private final ArrayList<IBodyDestroyListener> destroyListeners = new ArrayList<IBodyDestroyListener>();
	
	
	
	   /**
     * returns the id 
     * 
     * @return the id
     */
    abstract String getId();

    


	/**
	 * add a close lsitener 
	 * @param closeListener the close listener to add
	 */
	final void addCloseListener(IBodyCloseListener closeListener) {
		synchronized (closeListeners) {
			closeListeners.add(closeListener);
		}
	}
	

	/**
	 * add a destroy listener
	 * @param destroyListener the destroy listener to add 
	 */
	void addDestroyListener(IBodyDestroyListener destroyListener) {
        synchronized (destroyListeners) {
            destroyListeners.add(destroyListener);
        }
    }

	
    /**
	 * remove a close listener
	 * 
	 * @param closeListener a close listener
	 * @return true, if the listener is removed
	 */
	final boolean removeCloseListener(IBodyCloseListener closeListener) {
		synchronized (closeListeners) {
			return closeListeners.remove(closeListener);
		}
	}
	
	
	/**
     * remove a destroy listener
     * 
     * @param destroyListener a destroy listener
     * @return true, if the listener is removed
     */
    final boolean removeDestroyListener(IBodyDestroyListener destroyListner) {
        synchronized (destroyListeners) {
            return destroyListeners.remove(destroyListner);
        }
    }

	
	@SuppressWarnings("unchecked")
    protected final void callCloseListener() {
        ArrayList<IBodyCloseListener> closeListenersCopy = null;
        synchronized (closeListeners) {
            closeListenersCopy = (ArrayList<IBodyCloseListener>) closeListeners.clone();
        }
        
        for (IBodyCloseListener bodyCloseListener : closeListenersCopy) {
            removeCloseListener(bodyCloseListener);
            try {
                bodyCloseListener.onClose();
            } catch (IOException ioe) {
                if (LOG.isLoggable(Level.FINE)) {
                    LOG.fine("[" + getId() + "] error occured by calling " + bodyCloseListener + " " + ioe.toString());
                }
            }
        }
    }
	
	
	@SuppressWarnings("unchecked")
	protected final void callDestroyListener() {
	    ArrayList<IBodyDestroyListener> destroyListenersCopy = null;
	    synchronized (closeListeners) {
	        destroyListenersCopy = (ArrayList<IBodyDestroyListener>) destroyListeners.clone();
	    }
	        
	    for (IBodyDestroyListener bodyDestroyListener : destroyListenersCopy) {
	        removeDestroyListener(bodyDestroyListener);
	        try {
	            bodyDestroyListener.onDestroyed();
	        } catch (IOException ioe) {
	            if (LOG.isLoggable(Level.FINE)) {
	                LOG.fine("[" + getId() + "] error occured by calling " + bodyDestroyListener + " " + ioe.toString());
	            }
	        }
	    }
	}    
}
	    
